/*  Acceleration calculations for Greenspan/Barnes/Hut programs.	*/
/*  Source file:  fill.cs		Prototypes:  fill.hs		*/
/*  Uses  gbh.hs							*/

#include <stdio.h>
#include <stdlib.h>
#include <cscomm.h>
#include <float.h>
#include "gbh.hs"

#define Tm (*pTm)
#define Th (*pTh)

/*  Fill tree bottom-up with pseudo-particles.  Tree should already	*/
/*  have been constructed as by TreeTop (tree.cs).			*/
/*  Call from particle shape with all particles active.			*/

void Tfill(float:Tree Xp0[3], float:Tree Xp1[3],
	   float:Tree *pTm,   float:Tree *pTh    )
 { int:Tree rprt,		/* Number of children reporting */
       lwr, cmpl;		/* Lower-level, completed nodes	*/
   float:Tree Ca0[3], Cb0[3],	/* Box limits			*/
	      Ca1[3], Cb1[3];
   CMC_sendaddr_t:Tree ps;	/* Address of parent		*/
   CMC_sendaddr_t:Particles cs;	/* Address of leaf cell		*/
   int u, fvsz;
   
   /* Prepare tree */
   with ( Tree ) everywhere
    { ps = make_send_address(Tree, Parent);
      lwr = cmpl = 0;
      where ( Parent > 0 ) lwr = -1;
      rprt = 0;  Tm = 0.0;  Th = 0.0;		/* mass, length */
      for ( u = 0; u < 3; u++ )
       { Xp0[u] = Xp1[u] = 0.0;  Ca0[u] = Ca1[u] = FLT_MAX;
         Cb0[u] = Cb1[u] = -FLT_MAX;
       }
      /* Send data to leaves */
      with ( Particles ) everywhere
       { cs = make_send_address(Tree, cell);
         fvsz = 3*boolsizeof(float:Particles);
	 send(Xp0, &cs, X0, fvsz, CMC_no_field);
	 send(Xp1, &cs, X1, fvsz, CMC_no_field);
	 send(pTm, cs, gm, CMC_combiner_overwrite, CMC_no_field);
       }
      /* Tdetail(1, Xp0, Xp1, pTm, pTh);	    /*** Testing ***/
      cmpl = lwr & (Tm > 0.0f);		/* Leaves below level 1	*/
      /* Form boxes for leaves 	*/
      where ( Tm > 0.0f )		/* All leaves */
       for ( u = 0; u < 3; u++ )
        { Ca0[u] = Cb0[u] = Xp0[u];  Ca1[u] = Cb1[u] = Xp1[u];
	  Xp0[u] *= Tm;  Xp1[u] *= Tm;	/* For CoM calculations */
	}
      HT = 0;

      /* Loop to progagate particles up the tree:	*/
      /* cmpl identifies newly completed nodes.		*/
      /* rprt = number of children reporting so far.	*/
      /* Old completed nodes have  rprt = Chldrn+1 .	*/
      /* Other nodes have  rprt <= Chldrn .		*/
      /* Ca, Cb give box limits for nodes.		*/

      while ( |= cmpl )		/* Some nodes active	*/
       { HT++;
	 /* Tdetail(1, Xp0, Xp1, pTm, pTh);	    /*** Testing ***/
	 where ( cmpl )	/* Completed nodes	*/
	 { rprt++;
	   for ( u = 0; u < 3; u++ )
	    { send(&Xp0[u], ps, Xp0[u], CMC_combiner_add, CMC_no_field);
	      send(&Xp1[u], ps, Xp1[u], CMC_combiner_add, CMC_no_field);
	      send(&Ca0[u], ps, Ca0[u], CMC_combiner_min, CMC_no_field);
	      send(&Cb0[u], ps, Cb0[u], CMC_combiner_max, CMC_no_field);
	      send(&Ca1[u], ps, Ca1[u], CMC_combiner_min, CMC_no_field);
	      send(&Cb1[u], ps, Cb1[u], CMC_combiner_max, CMC_no_field);
	    }
	   send(pTm, ps, Tm, CMC_combiner_add, CMC_no_field);
	   send(&rprt, ps, (int:current) 1, CMC_combiner_add, CMC_no_field);
	 }
	cmpl = lwr & (rprt == Chldrn);
       }

      /* Finish calculating psuedo-particles, error parameters	*/
      Th = 0.0;  HT++;
      where ( Tm > 0.0f )
       { for ( u = 0; u < 3; u++ )
	  { Xp0[u] /= Tm;  Xp1[u]/= Tm;
	    Th >?= (Cb0[u] - Ca0[u]) >? (Cb1[u] - Ca1[u]);
	  }
	 where ( Chldrn == 0 ) Th = 0.0f;	/* Leaves */
       }
    }	/* end with ( Tree ) everywhere	*/
   /*  Restore exact coordinates in leaves	*/
   everywhere	/* in Particles */
    { send(Xp0, &cs, X0, fvsz, CMC_no_field);
      send(Xp1, &cs, X1, fvsz, CMC_no_field);
    }
 }
